#ifndef lint
static char Sccsid[] = "@(#)findrcds.c	4.3    DeltaDate 11/25/01    ExtrDate 12/10/01";
#endif

/*	FINDRCDS.C	*/
/*	This module is used to find records in the data base
**	matching the selection criteria.
**	The user may select multiple values of one field
**	connected by '&' or '|' which will be or'd and/or
**	multiple fields which will be and'd.
**	In other words, if field 1 is specified
**	as value1|value2 and field3 is specified as value3.
**	Records with field1==(value1 || value2) && field3==value3
**	will be selected. An external routine 'display', passed
**	as a parameter will be called to process the selected records.
**	Display must return 0 to go on to the next record, +1 to abort
**	the search and -1 to back up a record.
*/

#include <stdio.h>
#include <errno.h>
#include "ascii.h"
#include "cardfile.h"
#ifdef NEWS
extern int	errno;
#endif

extern	int	readonly;
static	int	match(char *rcd, struct Sdata *sp, char op, int fnum, int val_num);
static	char	*getslct();
	char	*getfld(),
		*pgetfld();
    
extern	char	*malloc();

static int field_match(char *re, char *line);


int
findrcds(struct Fdata *fields, char *dbname, int (*display)(), char *first)
{
    struct	Fdata	*fp;
    struct	Sdata	find_screen[MAXFLDS+1], *sp;
    int 	found;
    FILE	*ak_filep;
    FILE	*db_filep;
    char	filename[FNSIZE];
    diskptr	offset;
    char	*last =
		"Separate alternate values with a | for 'or' or a & for 'and'";
    int		knum, val_num;
    long	getkey();
    long	back_list[MAXKEYS];
    int		high;
    char	*keyval;
    char	*rcd;
    char	nextop;
    char	out_line[SWIDTH+1];
    int		i;
    int 	rc;
    
    fp = fields;
    sp = find_screen;
    while (fp->F_title[0] != '\0') {
	if (fp->F_key == 'N') {
	    ++fp;
	    continue;
	}
	sp->S_title = fp->F_title;
	sp->S_length = 2 * fp->F_length;
	sp->S_page = -1;
	sp->S_Lrow = -1;
	sp->S_Lcol = -1;
	sp->S_Drow = -1;
	sp->S_Dcol = -1;
	sp->S_Dfmt = "";
	sp->S_result = malloc((unsigned)sp->S_length+1);
	sp->S_dfault = 0;
	++fp;
	++sp;
    }
    sp->S_title = 0;
    sprintf(filename, "%s%s.db", datadir, dbname);
    if (readonly) {
	if ((db_filep = fopen(filename, "r")) == NULL) {
	    sprintf(out_line, "Unable to open %s, errno=%d\n", filename, errno);
	    msg(out_line);
	    return(1);
	}
    } else {
	if ((db_filep = fopen(filename, "r+")) == NULL) {
	    if (errno == EACCES) {
		strcpy(out_line, "You do not have permission to modify this database");
	    } else {
		sprintf(out_line, "Unable to open %s, errno=%d\n", filename,
		    errno);
	    }
	    msg(out_line);
	    return(1);
	}
    }
    rcd = malloc(DBSIZE);
    while (screen(first, find_screen, last, 0, FALSE, 0) != 0) {
	sp = find_screen;
	knum = -1;
	found = 0;
	while (sp->S_title) {		/* loop on key field */
	    ++knum;
	    if (*(sp->S_result) == '\0') {	/* was key entered */
		++sp;
		continue;
	    }
	    sprintf(filename, "%s%s.ak%d", datadir, dbname, knum);
	    if ((ak_filep = fopen(filename, "r")) == NULL) {
		msg("Unable to open alternate key file");
		getout();
	    }
	    high = 0;
	    val_num = 0;
	    /* loop for each key for this field */
	    while ((keyval = getslct(sp->S_result, "&|", &nextop, ++val_num)) != NULL) {
		/* loop getting records for this value */
		while ((offset = getkey(ak_filep, keyval)) >= 0L) {
		    keyval = 0;
		    fseek(db_filep, offset, 0);
		    if (fgets(rcd, DBSIZE, db_filep) == NULL) {
			msg("Error reading DB file");
			getout();
		    }
		    if (*rcd == 'D') {
			continue;
		    }
		    ++found;
		    if (match(rcd, sp, nextop, knum, val_num)) {
			/* save offset for reverse display */
			back_list[high] = offset;
			if (high < MAXKEYS)
			    ++high;
			if ((rc = (*display)(fields, rcd, db_filep, dbname)) == 1)
			    break;
			if (rc == -1) {	/* reverse */
			    /* DJL - back up list */
			    for (i = high - 2 ; i >= 0 && rc == -1 ; --i) {
				fseek(db_filep, back_list[i], 0);
				fgets(rcd, DBSIZE, db_filep);
				if (*rcd == 'D')
				    continue;
				rc = (*display)(fields, rcd, db_filep, dbname);
			    }
			    if (rc == 1)
				break;	/* escape from inside reverse loop */
			}
		    }
		}
		/* eat '&' and find next '|' */
		while (nextop == '&') {
		    if (getslct(sp->S_result, "&|", &nextop, ++val_num) == NULL)
			break;
		}
	    }
	    fclose(ak_filep);
	    break;
	}
	if (found == 0) {
	    msg("No records found");
	}
    }
    fclose(db_filep);
    free(rcd);
    sp = find_screen;
    while (sp->S_title) {
	free(sp->S_result);
	++sp;
    }
    return(0);
}


/* identical to getfield except separator */
/* is returned thru sepval */
static char *
getslct(char *string, char *seps, char *sepval, int field)
{
    register char	*p, *r;
    static   char	savept[128];
    
    p = string;
    if (p == NULL || *p == '\0')
	return(NULL);
    while (--field > 0) {
	if ((r = strpbrk(p, seps)) == NULL) {	/* move past token */
	    *sepval = '\0';
	    return(NULL);
	} else {
	    p = r+1;				/* move past separator */
	}
    }
    strncpy(savept, p, 128);
    if ((r = strpbrk(savept, seps)) == NULL) {
	*sepval = '\0';
    } else {
	*sepval = *r;
	*r = '\0';
    }
    return(savept);
}


/*
 * This record is a possible match, see if all selection criteria match
 */
static int
match(char *rcd, struct Sdata *sp, char op, int fnum, int val_num)
{
    char	*field;

    for ( ; sp->S_title != NULL ; ++sp , ++fnum) {    /* loop for each field */
	if (sp->S_result[0] == '\0')
	    continue;		/* no value, skip field */
	field = pgetfld(rcd, fnum+1, ':');
	if (field_match(sp->S_result, field) == FALSE)
	    return(FALSE);
    }
    return(TRUE);
}


static int
field_match(char *re, char *line)
{
    int		val_num;
    char	*value;
    char	op;

    val_num = 0;
    while ((value = getslct(re, "&|", &op, ++val_num)) != NULL) {
	if (keymatch(line, value) == FALSE) {
	    if (op != '|') {
		return(FALSE);
	    }
	} else {
	    if (op != '&') {
		return(TRUE);
	    }
	}
    }
    return(TRUE);
}

